use std::{
    env, fs,
    io::{self, Read, Write},
    path::Path,
    process,
    time::Instant,
};

#[cfg(not(windows))]
use std::os::unix::fs::PermissionsExt;

use clap::Parser;
use crypto::util::fixed_time_eq;
use env_logger::Builder;
use log::LevelFilter;

mod util;
mod vitrify;

#[macro_use]
extern crate log;

const VITRIFYR_FILE_EXT: &str = ".vfyd";
const VITRIFYR_KEY_ENV_VAR: &str = "VITRIFYR_KEY";

#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Args {
    /// Input file to (de)vitrify (read from stdin by default)
    #[arg(short, long, default_value_t = String::from("-"))]
    input: String,

    /// Output directory to save (de)vitrified file to (current directory by default)
    #[arg(short, long, default_value_t = String::from("."))]
    output: String,

    /// Devitrify input file instead of vitrifying it
    #[arg(short, long, default_value_t = false)]
    devitrify: bool,

    /// AES (256-CBC) encryption/decryption key (optional)
    ///
    /// '0x' and '0X' prefixes will cause argument to be decoded as hexadecimal. You can also
    /// provide key through [`VITRIFYR_KEY`] environment variable.
    #[arg(short, long)]
    key: Option<String>,

    /// Chunk size used to process input file (128KB by default)
    #[arg(long, default_value_t = 131070)]
    chunk_size: usize,

    /// xz compression level preset (0 to 9, "faster" to "better" compression)
    #[arg(long, default_value_t = 6)]
    compression_level: u8,

    /// Disable vitrifyr integrity feature
    ///
    /// When vitrifying, input file content digest is stored as output file name (`${DIGEST}.vfyd`).
    /// When devitrifying, input file name is parsed and inferred digest is checked against output
    /// file content.
    /// Specifying this option disable both features.
    #[arg(long, default_value_t = false)]
    skip_integrity: bool,

    /// Do not write final output file path to stdout before exiting
    #[clap(short, long)]
    quiet: bool,

    /// Show informational logs
    #[clap(short, long)]
    verbose: bool,

    /// Show debugging logs
    #[clap(long)]
    debug: bool,
}

fn main() {
    let args = Args::parse();

    let mut builder = Builder::new();
    builder.filter_level(if args.debug {
        LevelFilter::Debug
    } else if args.verbose {
        LevelFilter::Info
    } else {
        LevelFilter::Warn
    });
    builder.parse_default_env();
    builder.init();

    if !(0..10).contains(&args.compression_level) {
        error!("unsupported compression level : {}", args.compression_level);
        process::exit(2);
    }

    let mut input_file: Box<dyn Read> = if args.input == "-" {
        info!("input file will be read from stdin");
        Box::new(io::stdin())
    } else {
        match fs::File::open(&args.input) {
            Ok(file) => {
                match fs::metadata(&args.input) {
                    Ok(metadata) => info!("input file is {}-byte long", metadata.len()),
                    Err(error) => warn!("couldn't read input file metadata : {}", error),
                }

                Box::new(file)
            }
            Err(error) => {
                error!("couldn't open {} for reading : {}", args.input, error);
                process::exit(1);
            }
        }
    };

    let mut output_path = Path::new(&args.output).join(
        util::random_token(8)
            + (if !args.devitrify {
                VITRIFYR_FILE_EXT
            } else {
                ""
            }),
    );

    // setup signal handler
    {
        // see <https://github.com/rust-lang/rfcs/issues/2407>
        let output_path = output_path.clone();
        ctrlc::set_handler(move || {
            debug!("caught termination signal, removing temporary file before exiting...");
            util::remove_file(&output_path);
            process::exit(2);
        })
        .expect("couldn't setup signal handler");
    }

    let mut output_file: Box<dyn Write> = match fs::File::create(&output_path) {
        Ok(file) => {
            #[cfg(not(windows))]
            if let Err(error) = file.set_permissions(fs::Permissions::from_mode(0o640)) {
                warn!("couldn't set output file permissions : {}", error);
            }

            debug!("successfully created output file {:?}", output_path);
            Box::new(file)
        }
        Err(error) => {
            error!("couldn't open {} for writing : {}", args.output, error);
            process::exit(1);
        }
    };

    // fetch encryption key from environment or CLI (if set)
    let input_key: Option<String> = match env::var(VITRIFYR_KEY_ENV_VAR) {
        Ok(value) => Some(value),
        Err(error) => {
            // allow environment variable to be missing
            if error != env::VarError::NotPresent {
                debug!(
                    "couldn't interpret {} environment variable",
                    VITRIFYR_KEY_ENV_VAR
                );
                process::exit(1);
            }

            args.key
        }
    };

    // convert encryption key as bytes ready for AES-256
    let key: Option<[u8; vitrify::AES_KEY_SIZE]> = match input_key {
        Some(k) => {
            // decode encryption key as hexadecimal (if needed)
            let mut key = if k.to_lowercase().starts_with("0x") {
                if k.len() % 2 != 0 {
                    error!(
                        "cannot decode hexadecimal from odd-length encryption key : {}",
                        k
                    );
                    process::exit(2);
                }

                match util::hex_decode(&k[2..]) {
                    Ok(k) => k,
                    Err(error) => {
                        error!(
                            "cannot decode hexadecimal encryption key from {} : {}",
                            k, error
                        );
                        process::exit(2);
                    }
                }
            } else {
                k.into_bytes()
            };

            if key.len() > vitrify::AES_KEY_SIZE {
                warn!(
                    "encryption key is too long and will be truncated to {} bytes",
                    vitrify::AES_KEY_SIZE
                );
            }

            key.resize(vitrify::AES_KEY_SIZE, 0);

            debug!("encryption key is set to : {:?}", key);
            Some(key.try_into().unwrap())
        }
        None => None,
    };

    let now = Instant::now();

    if args.devitrify {
        // BEGIN DEVITRIFICATION
        let mut input_digest =
            match vitrify::decode(&mut input_file, &mut output_file, args.chunk_size, &key) {
                Ok(digest) => digest,
                Err(error) => {
                    error!("devitrifying input file failed : {}", error);

                    util::remove_file(&output_path);
                    process::exit(1);
                }
            };

        debug!(
            "successfully devitrified input file (digest={}, size={}) to {:?} in {:?}",
            input_digest.result_str(),
            match fs::metadata(&output_path) {
                Ok(metadata) => metadata.len() as i64,
                Err(error) => {
                    warn!("couldn't read output file metadata : {}", error);

                    -1
                }
            },
            output_path,
            now.elapsed()
        );

        // if not disabled, check output file content digest against its hexadecimal coming from
        // input filename
        if !args.skip_integrity {
            match Path::new(&args.input).file_stem() {
                None => {
                    if args.input == "-" {
                        debug!(
                            "cannot check input file integrity when reading from stdin, ignoring..."
                        );
                    } else {
                        warn!(
                            "couldn't infer input file stem portion from {}, skipping integrity check...",
                            args.input
                        );
                    }
                }
                Some(file_stem) => {
                    let hex_digest = file_stem.to_string_lossy();
                    if hex_digest.len() % 2 != 0 {
                        error!(
                            "cannot decode hexadecimal from odd-length input filename : {}",
                            hex_digest
                        );
                        process::exit(2);
                    }

                    match util::hex_decode(&hex_digest) {
                        Err(error) => warn!(
                            "couldn't decode hexadecimal from input filename {} ({}), skipping integrity check...",
                            hex_digest, error
                        ),
                        Ok(d) => {
                            let mut digest_result = vec![0; input_digest.output_bytes()];
                            input_digest.result(&mut digest_result);
                            if !fixed_time_eq(&digest_result, &d) {
                                error!(
                                    "output file content didn't pass integrity check (digest={:?})",
                                    input_digest.result_str()
                                );
                                process::exit(1);
                            }

                            info!(
                                "successfully checked output file content integrity : {}", hex_digest
                            );
                        }
                    }
                }
            }
        }
        // END DEVITRIFICATION
    } else {
        // BEGIN VITRIFICATION
        let input_hex_digest = match vitrify::encode(
            &mut input_file,
            &mut output_file,
            args.chunk_size,
            args.compression_level,
            &key,
        ) {
            Ok(mut digest) => digest.result_str(),
            Err(error) => {
                error!("vitrifying input file failed : {}", error);

                util::remove_file(&output_path);
                process::exit(1);
            }
        };

        debug!(
            "successfully vitrified input file (digest={}, size={}) to {:?} in {:?}",
            input_hex_digest,
            match fs::metadata(&output_path) {
                Ok(metadata) => metadata.len() as i64,
                Err(error) => {
                    warn!("couldn't read output file metadata : {}", error);

                    -1
                }
            },
            output_path,
            now.elapsed()
        );

        // if not disabled, store input file content digest as hexadecimal filename
        if !args.skip_integrity {
            let new_output_path =
                Path::new(&args.output).join(input_hex_digest + VITRIFYR_FILE_EXT);

            if let Err(error) = fs::rename(&output_path, &new_output_path) {
                warn!(
                    "couldn't rename output file to {:?} : {}",
                    new_output_path, error
                );
            } else {
                debug!(
                    "successfully renamed output file from {:?} to {:?}",
                    output_path, new_output_path
                );
                output_path = new_output_path;
            }
        }
        // END VITRIFICATION
    }

    // write final output file path to stdout so external scripts can easily retrieve its location
    if !args.quiet {
        println!("{}", output_path.to_string_lossy());
    }
}
