use std::io::{self, Read, Write};

use aesstream::{AesReader, AesWriter};
use base64::{
    engine::general_purpose::STANDARD, read::DecoderReader as B64DecoderReader,
    write::EncoderWriter as B64EncoderWriter,
};
use crypto::{
    aessafe::{AesSafe256Decryptor, AesSafe256Encryptor},
    digest::Digest,
    sha2::Sha256,
};
use liblzma::{read::XzDecoder, write::XzEncoder};

pub const AES_KEY_SIZE: usize = 256 / 8;

/// Perform file vitrification (base64 encoding + xz compression) and return input file content
/// digest as `crypto::digest::Digest` instance.
/// If an encryption `key` is specified, output bytes are also encrypted using AES-256-CBC.
pub fn encode(
    input_file: &mut Box<dyn Read>,
    output_file: &mut Box<dyn Write>,
    chunk_size: usize,
    compression_level: u8,
    key: &Option<[u8; AES_KEY_SIZE]>,
) -> Result<Box<dyn Digest>, io::Error> {
    let mut digest = Sha256::new();

    // if a key has been provided, prepare an AES-256-CBC encrypting stream writing to output file
    let output_writer: Box<dyn Write> = match key {
        Some(k) => Box::new(AesWriter::new(output_file, AesSafe256Encryptor::new(k))?),
        None => Box::new(output_file),
    };

    // create a xz encoder with above writer as output
    let xz_encoder = XzEncoder::new_parallel(output_writer, compression_level as u32);

    // create a base64 encoder with above xz encoder as output
    let mut b64_encoder = B64EncoderWriter::new(xz_encoder, &STANDARD);

    let mut buffer = vec![0; chunk_size];
    loop {
        let nb_bytes = input_file.read(&mut buffer)?;
        if nb_bytes == 0 {
            break;
        }

        b64_encoder.write_all(&buffer[..nb_bytes])?;

        // update digest context with processed bytes
        digest.input(&buffer[..nb_bytes]);
    }

    Ok(Box::new(digest))
}

/// Perform file devitrification (xz decompression + base64 decoding) and return output file
/// content digest as `crypto::digest::Digest` instance.
/// If a decryption `key` is specified, input bytes are first decrypted using AES-256-CBC.
pub fn decode(
    input_file: &mut Box<dyn Read>,
    output_file: &mut Box<dyn Write>,
    chunk_size: usize,
    key: &Option<[u8; AES_KEY_SIZE]>,
) -> Result<Box<dyn Digest>, io::Error> {
    let mut digest = Sha256::new();

    // if a key has been provided, prepare an AES-256-CBC decrypting stream reading from input file
    let input_reader: Box<dyn Read> = match key {
        Some(k) => Box::new(AesReader::new(input_file, AesSafe256Decryptor::new(k))?),
        None => Box::new(input_file),
    };

    // create a xz decoder with above reader as input
    let xz_decoder = XzDecoder::new_parallel(input_reader);

    // create a base64 decoder with above xz decoder as input
    let mut b64_decoder = B64DecoderReader::new(xz_decoder, &STANDARD);

    let mut buffer = vec![0; chunk_size];
    loop {
        let nb_bytes = b64_decoder.read(&mut buffer)?;
        if nb_bytes == 0 {
            break;
        }

        output_file.write_all(&buffer[..nb_bytes])?;

        // update digest context with processed bytes
        digest.input(&buffer[..nb_bytes]);
    }

    Ok(Box::new(digest))
}
